/*******************************************************************************
 * Copyright (c) 2022, Jan Koester jan.koester@gmx.net
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * Neither the name of the <organization> nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *******************************************************************************/

#include "stdint.h"

#include <bits/sysbits.h>
#include <bits/syscall.h>

#include "mutex_internal.h"

#include <config.h>

#include <sys/mman.h>


namespace sys {
    namespace __internal {
        class allocator {
        public:
            static allocator& getInstance() noexcept{
                static allocator instance;
                return instance;
            }

            void *alloc(unsigned long size) noexcept{
                unsigned long blksize=size+(sizeof(heap));

                heap *block=(heap*)mmap(0,blksize, PROT_READ | PROT_WRITE,MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);

                if((long)block<0){
                    return nullptr;
                }

                zero(block,sizeof(heap));
                block->_size=size;
                block->_total=blksize;
                block->_block=(char*)block+(sizeof(heap));
                block->_prevheap=_lastheap;
                uint32_t exp = 0;
                uint32_t des = 1;
                __lock_mutex(_memlock,exp,des);
                _lastheap=block;
                __unlock_mutex(_memlock,des,exp);
                return block->_block;
            }

            void free(void* ptr) noexcept{
                uint32_t exp = 0;
                uint32_t des = 1;
                __lock_mutex(_memlock,exp,des);
                heap *nextheap=nullptr;
                for(heap *curheap=_lastheap; curheap; curheap=curheap->_prevheap){
                    if(curheap->_block==ptr){
                        if(_lastheap==curheap)
                            _lastheap=curheap->_prevheap;
                        if(nextheap)
                            nextheap->_prevheap=curheap->_prevheap;
                        munmap(curheap,curheap->_total);
                        __unlock_mutex(_memlock,des,exp);
                        return;
                    }
                    nextheap=curheap;
                }
                __unlock_mutex(_memlock,des,exp);
            }

            allocator(){
                uint32_t exp = 1;
                uint32_t des = 0;
                _memlock[0]=__init_mutex(exp,des);
                _lastheap=nullptr;
            };

            ~allocator(){
                for(heap *curheap=_lastheap; curheap; curheap=curheap->_prevheap){
                    free(curheap);
                }
                uint32_t exp = 0;
                uint32_t des = 1;
                __destroy_mutex(_memlock,exp,des);
            };
        private:

            class heap {
            public:
                void          *_block;
            private:
                ~heap();
                unsigned long  _size;
                unsigned long  _total;
                heap          *_prevheap;
                friend class allocator;
            };


            void zero(void *s, unsigned n){
                char *str = (char *)s;
                while (n--)
                    str[n] = 0;
            };

            heap     *_lastheap;
            uint32_t  _memlock[];
        };
    };
};

void *operator new(unsigned long size){
    return sys::__internal::allocator::getInstance().alloc(size);
}

void *operator new[](unsigned long size){
    return sys::__internal::allocator::getInstance().alloc(size);
}

void operator delete (void* ptr) throw(){
    sys::__internal::allocator::getInstance().free(ptr);
}

void operator delete[] (void* ptr) throw(){
    sys::__internal::allocator::getInstance().free(ptr);
}

void operator delete(void* ptr,unsigned long amount){
    sys::__internal::allocator::getInstance().free(ptr);
}

void operator delete[](void* ptr,unsigned long amount){
    sys::__internal::allocator::getInstance().free(ptr);
}
