/*******************************************************************************
 * Copyright (c) 2023, Jan Koester jan.koester@gmx.net
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * Neither the name of the <organization> nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *******************************************************************************/

#include <errno.h>
#include <string.h>

#include <bits/sysbits.h>
#include <bits/syscall.h>

#include <sys/ressource.h>

#define MIN(a, b) ((a)<(b) ? (a) : (b))
#define FIX(x) do{ if ((x)>=SYSCALL_RLIM_INFINITY) (x)=RLIM_INFINITY; }while(0)

#ifndef SYSCALL_RLIM_INFINITY
#define SYSCALL_RLIM_INFINITY (~0ULL)
#endif

struct ctx {
	unsigned long lim[2];
	int res;
	int err;
};

extern void __synccall(...);

extern "C" {
	int getrusage(int who, struct rusage* ru) {
		int r;
#ifdef __NR_getrusage_time64
		long long kru64[18];
		r = __syscall3(__NR_getrusage_time64, (unsigned long) who, kru64);
		if (!r) {
			ru->ru_utime = (struct timeval)
			{ .tv_sec = kru64[0], .tv_usec = kru64[1] };
			ru->ru_stime = (struct timeval)
			{ .tv_sec = kru64[2], .tv_usec = kru64[3] };
			char* slots = (char*)&ru->ru_maxrss;
			for (int i = 0; i < 14; i++)
				*(long*)(slots + i * sizeof(long)) = kru64[4 + i];
		}
		if (__NR_getrusage_time64 == __NR_getrusage || r != -ENOSYS)
			return syscall_ret(r);
#endif
		char* dest = (char*)&ru->ru_maxrss - 4 * sizeof(long);
		r = syscall2(__NR_getrusage, who, (unsigned long)dest);
		if (!r && sizeof(time_t) > sizeof(long)) {
			long kru[4];
			memcpy(kru, dest, 4 * sizeof(long));
			ru->ru_utime = (struct timeval)
			{ .tv_sec = kru[0], .tv_usec = kru[1] };
			ru->ru_stime = (struct timeval)
			{ .tv_sec = kru[2], .tv_usec = kru[3] };
		}
		return syscall_ret(r);
	}

	int getrlimit(int resource, struct rlimit* rlim){
		int ret = syscall4(__NR_prlimit64, 0, resource, 0, (unsigned long) rlim);
		if (!ret) {
			FIX(rlim->rlim_cur);
			FIX(rlim->rlim_max);
		}
#ifdef __NR_getrlimit
		unsigned long k_rlim[2];
		if (!ret || errno != ENOSYS)
			return ret;
		if (syscall2(__NR_getrlimit, resource, (unsigned long)k_rlim) < 0)
			return -1;
		rlim->rlim_cur = k_rlim[0] == -1UL ? RLIM_INFINITY : k_rlim[0];
		rlim->rlim_max = k_rlim[1] == -1UL ? RLIM_INFINITY : k_rlim[1];
		FIX(rlim->rlim_cur);
		FIX(rlim->rlim_max);
		return 0;
#else
		return ret;
#endif
	}

#ifdef __NR_setrlimit
	static void do_setrlimit(void* p)
	{
		struct ctx* c = (struct ctx*) p;
		if (c->err > 0) return;
		c->err = -syscall2(__NR_setrlimit,(unsigned long) c->res, (unsigned long)c->lim);
	}
#endif

	int setrlimit(int resource, const struct rlimit* rlim){
			struct rlimit tmp;
			if (SYSCALL_RLIM_INFINITY != RLIM_INFINITY) {
				tmp = *rlim;
				FIX(tmp.rlim_cur);
				FIX(tmp.rlim_max);
				rlim = &tmp;
			}
			int ret = syscall4(__NR_prlimit64, 0, resource, (unsigned long) rlim, 0);
#ifdef __NR_setrlimit
			if (ret != -ENOSYS) return syscall_ret(ret);

			struct ctx c = {
				.lim[0] = MIN(rlim->rlim_cur, MIN(-1UL, SYSCALL_RLIM_INFINITY)),
				.lim[1] = MIN(rlim->rlim_max, MIN(-1UL, SYSCALL_RLIM_INFINITY)),
				.res = resource, .err = -1
			};
			__synccall(do_setrlimit, &c);
			if (c.err) {
				if (c.err > 0) errno = c.err;
				return -1;
			}
			return 0;
#else
			return syscall_ret(ret);
#endif
	}


};