/*******************************************************************************
 * Copyright (c) 2022, Jan Koester jan.koester@gmx.net
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * Neither the name of the <organization> nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *******************************************************************************/

#include <string.h>
#include <errno.h>
#include <signal.h>

#include <systempp.h>
#include <ksigaction.h>
#include <atomic.h>

#include "thread_internal.h"

namespace sys {
    namespace __internal {
        static const unsigned long app_mask[] = {
#if ULONG_MAX == 0xffffffff
#if _NSIG == 65
            0x7fffffff, 0xfffffffc
#else
            0x7fffffff, 0xfffffffc, -1UL, -1UL
#endif
#else
#if _NSIG == 65
            0xfffffffc7fffffff
#else
            0xfffffffc7fffffff, -1UL
#endif
#endif
        };

        static const unsigned long all_mask[] = {
#if ULONG_MAX == 0xffffffff && _NSIG > 65
            - 1UL, -1UL, -1UL, -1UL
#elif ULONG_MAX == 0xffffffff || _NSIG > 65
            - 1UL, -1UL
#else
            - 1UL
#endif
        };

        void __block_all_sigs(void* set){
            syscall4(__NR_rt_sigprocmask, SIG_BLOCK, (unsigned long)&all_mask, (unsigned long) set, _NSIG / 8);
        }

        void __block_app_sigs(void* set){
            syscall4(__NR_rt_sigprocmask, SIG_BLOCK, (unsigned long)&app_mask, (unsigned long)set, _NSIG / 8);
        }

        void __restore_sigs(void* set){
            syscall4(__NR_rt_sigprocmask, SIG_SETMASK, (unsigned long) set, 0, _NSIG / 8);
        }

#define IS32BIT(x) !((x)+0x80000000ULL>>32)
#define CLAMP(x) (int)(IS32BIT(x) ? (x) : 0x7fffffffU+((0ULL+(x))>>63))

        static int do_sigtimedwait(const sigset_t* __restrict__ mask, siginfo_t* __restrict__ si, const struct timespec* __restrict__ ts)
        {
#ifdef SYS_rt_sigtimedwait_time64
            time_t s = ts ? ts->tv_sec : 0;
            long ns = ts ? ts->tv_nsec : 0;
            int r = -ENOSYS;
            if (SYS_rt_sigtimedwait == SYS_rt_sigtimedwait_time64 || !IS32BIT(s))
                r = syscall4(SYS_rt_sigtimedwait_time64, (unsigned long)mask, (unsigned long)si,
                    ts ? ((long long[]) { s, ns }) : 0, (unsigned long)_NSIG / 8);
            if (SYS_rt_sigtimedwait == SYS_rt_sigtimedwait_time64 || r != -ENOSYS)
                return r;
            return syscall4(SYS_rt_sigtimedwait, (unsigned long) mask, (unsigned long) si,
                ts ? ((long[]) { CLAMP(s), ns }) : 0, (unsigned long) _NSIG / 8);;
#else
            return syscall4(__NR_rt_sigtimedwait,(unsigned long) mask, (unsigned long)si, (unsigned long)ts, (unsigned long)_NSIG / 8);
#endif
        }
    };
};

extern "C" {

    int __libc_current_sigrtmin(void) {
        return 0;
    }

    int __libc_current_sigrtmax(void) {
        return 0;
    }

    int kill(pid_t, int) {
        return 0;
    }

    int sigemptyset(sigset_t*) {
        return 0;
    }

    int sigfillset(sigset_t*) {
        return 0;
    }

    int sigaddset(sigset_t*, int) {
        return 0;
    }

    int sigdelset(sigset_t*, int) {
        return 0;
    }

    int sigismember(const sigset_t*, int) {
        return 0;
    }

    int sigprocmask(int, const sigset_t* __restrict__ sa, sigset_t* __restrict__ se) {
        return 0;
    }

    int sigsuspend(const sigset_t*) {
        return 0;
    }

    int sigaction(int sig, const struct sigaction* __restrict__ sa, struct sigaction* __restrict__ old) {
        unsigned long set[_NSIG / (8 * sizeof(long))];

        if (sig - 32U < 3 || sig - 1U >= _NSIG - 1) {
            errno = EINVAL;
            return -1;
        }

        /* Doing anything with the disposition of SIGABRT requires a lock,
         * so that it cannot be changed while abort is terminating the
         * process and so any change made by abort can't be observed. */
        if (sig == SIGABRT) {
            sys::__internal::__block_all_sigs(&set);
            sys::__internal::thread::getins().lock(__abort_lock);
        }
        int r = ksigaction(sig, sa, old);
        if (sig == SIGABRT) {
            sys::__internal::thread::getins().unlock(__abort_lock);
            sys::__internal::__restore_sigs(&set);
        }
        return r;
    }

    int sigpending(sigset_t*) {
        return 0;
    }

    int sigwait(const sigset_t* __restrict__ mask, int* __restrict__ sig) {
        siginfo_t si;
        if (sigtimedwait(mask, &si, 0) < 0)
            return -1;
        *sig = si.si_signo;
        return 0;
    }

    int sigwaitinfo(const sigset_t* __restrict, siginfo_t* __restrict) {
        return 0;
    }

    int sigtimedwait(const sigset_t* __restrict__ mask, siginfo_t* __restrict__ si, const struct timespec* __restrict__ timeout)
    {
        int ret;
        do ret = sys::__internal::do_sigtimedwait(mask, si, timeout);
        while (ret == -EINTR);
        return syscall_ret(ret);
    }

    int sigqueue(pid_t, int, const union sigval) {
        return 0;
    }

    int pthread_sigmask(int, const sigset_t* __restrict, sigset_t* __restrict) {
        return 0;
    }

    int pthread_kill(pthread_t, int) {
        return 0;
    }

    void psiginfo(const siginfo_t*, const char*) {
    }

    void psignal(int, const char*) {
    }

    int killpg(pid_t, int) {
        return 0;
    }

    int sigaltstack(const stack_t* __restrict, stack_t* __restrict) {
        return 0;
    }

    int sighold(int) {
        return 0;
    }

    int sigignore(int) {
        return 0;
    }

    int siginterrupt(int, int) {
        return 0;
    }

    int sigpause(int) {
        return 0;
    }

    int sigrelse(int) {
        return 0;
    }

    void (*sigset(int, void (*)(int)))(int){
        return nullptr;
    }

    int raise(int sig){
        sigset_t set;
        sys::__internal::__block_app_sigs(&set);
      //   int ret = syscall(__NR_tkill, __pthread_self()->tid, sig);
        sys::__internal::__restore_sigs(&set);
     //   return ret;
    };  

    void (*signal(int, void (*)(int)))(int){
        return nullptr;
    }
};
