/*******************************************************************************
Copyright (c) 2022, Jan Koester jan.koester@gmx.net
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
	* Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.
	* Neither the name of the <organization> nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

#include <errno.h>
#include <limits.h>
#include <sched.h>

#include <sys/mman.h>
#include <bits/syscall.h> 
#include <bits/sysbits.h>
#include <pthread.h>

#include "atomic.h"
#include "systempp.h"
#include "thread_internal.h"
#include "mutex_internal.h"

#include <stdlib.h>
#include <sched.h>

#include "console.h"

#define DEFAULT_STACKSIZE (4096 * 4096)

#define ARCH_SET_GS 0x1001
#define ARCH_SET_FS 0x1002
#define ARCH_GET_FS 0x1003
#define ARCH_GET_GS 0x1004

struct sys::__internal::pthread_key_struct sys::__internal::__pthread_keys[128];
struct sys::__internal::pthread_key_data sys::__internal::specific_1stblock[32];

extern "C" {
    int __clone(void*(*fn)(void*),void*stack,int flags,void*arg,...);
}
   
sys::__internal::thread& sys::__internal::thread::getins(){
    static thread ins;
    return ins;
};
    
void sys::__internal::thread::addins(sys::thread *thrd,size_t tid){
    lock(_mutex);
    if(_firstins){
        _lastins->nextins=new ins;
        _lastins=_lastins->nextins;
    }else{
        _firstins = new ins;
        _lastins = _firstins;
    }
    _lastins->tid=tid;
    _lastins->threadins=thrd;
    unlock(_mutex);
};
    
void sys::__internal::thread::delins(sys::thread *thrd){
        
};

sys::thread* sys::__internal::thread::self(){
     lock(_mutex);
    for (sys::__internal::thread::ins* myins = sys::__internal::thread::getins()._firstins; myins; myins = myins->nextins) {
        if (syscall0(__NR_gettid) == myins->tid) {
            unlock(_mutex);
            return myins->threadins;
        }
    }
    unlock(_mutex);
    return nullptr;
}

void* sys::__internal::thread::getDesc(size_t tid){
    lock(_mutex);
    for(ins *curins=_firstins; curins; curins=curins->nextins){
        cout << tid << ":"<< curins->tid << "addr: " << (size_t) &curins->threadins << endl;
        if (curins->tid == tid) {
            unlock(_mutex);
            return ((sys::thread*)curins->threadins)->getDesc();
        }
    }
    unlock(_mutex);
    return nullptr;
};

int sys::__internal::thread::getPid(size_t tid){
    for(ins *curins=_firstins; curins; curins=curins->nextins){
        cout << tid << ":"<< curins->tid << "addr: " << (size_t) &curins->threadins << endl;
        if (curins->tid == tid) {
            unlock(_mutex);
            return ((sys::thread*)curins->threadins)->getPid();
        }
    }
    unlock(_mutex);
    return -1;
};

void sys::__internal::thread::lock(volatile int* l) {
    uint32_t des = 1;
    uint32_t exp = 0;
    __lock_mutex((uint32_t*)l,exp, des);
}

void sys::__internal::thread::unlock(volatile int* l) {
    uint32_t des = 0;
    uint32_t exp = 1;
    __internal::__unlock_mutex((uint32_t*)l,exp, des);
}

sys::__internal::thread::ins::ins(){
    nextins=nullptr;
}

sys::__internal::thread::ins::~ins(){
    delete nextins;
}

sys::__internal::thread::thread(){
    _firstins=nullptr;
    _lastins=nullptr;
    uint32_t exp = 1;
    uint32_t des = 0;
    _mutex[0]=__init_mutex(exp, des);
}

sys::__internal::thread::~thread(){
    delete _firstins;
    uint32_t exp = 0;
    uint32_t des = 1;
    __destroy_mutex((uint32_t*)_mutex,exp, des);
}

sys::thread::thread(void *args,threadattr *attr){
    _nextThread=nullptr;

    if(!attr){
        attr=new threadattr;
        _selfattr=true;
    }else{
        _selfattr=false;
    }

    attr->_finalize();

    const unsigned long clone_flags = CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND
                                        | CLONE_THREAD | CLONE_SYSVSEM | CLONE_SETTLS
                                        | CLONE_PARENT_SETTID | CLONE_CHILD_CLEARTID;
    __Args.ins=this;
    __Args.args=args;

    _joinable=false;

    int _tls=syscall2(__NR_arch_prctl,ARCH_SET_GS,(unsigned long)&_udesc);

    if(_tls<0){
        SystemException excep;
        throw excep[SystemException::Error] << "Clone failed can't allocate thread local space !";
    }

    int ret = __clone((void* (*)(void*))&this->_start,attr->_stack+attr->_stacksize,clone_flags,&__Args,&_ParentPid,&_ThreadPid,_tls);

    if(ret <0){
        SystemException excep;
        throw excep[SystemException::Error] << "Clone failed can't start thread !";
    }

    if(ret >0)
        sys::__internal::thread::getins().addins(this,ret);

    sys::__internal::systempp::Threaded = true;
}

void* sys::thread::_start(void *ins) noexcept{
    try {
        sys::thread *mythread=(sys::thread*)((__ThreadArguments*)ins)->ins;
        mythread->_Retval = mythread->run(((__ThreadArguments*)ins)->args);
        mythread->_joinable=true;
        return mythread->_Retval;
    }catch(SystemException &e){
        return (void*)&e;
    }
}

extern "C" {
    pthread_t pthread_self(void) {
        for (sys::__internal::thread::ins* myins = sys::__internal::thread::getins()._firstins; myins; myins = myins->nextins) {
            if (syscall0(__NR_gettid) == myins->tid) {
                return myins->threadins;
            }
        }
        return nullptr;
    }

    int pthread_key_create(pthread_key_t* key, void (*destructor)(void*)) {
        /* Find a slot in __pthread_kyes which is unused.  */
        for (size_t cnt = 0; cnt < PTHREAD_KEYS_MAX; ++cnt)
        {
            uintptr_t seq = sys::__internal::__pthread_keys[cnt].seq;

            if ((((seq) & 1) == 0) && (((uintptr_t)(seq)) < ((uintptr_t)((seq)+2)))
                /* We found an unused slot.  Try to allocate it.  */
                && !__atomic_compare_exchange(&sys::__internal::__pthread_keys[cnt].seq,
                    &seq+1,&seq,false, __ATOMIC_ACQ_REL, __ATOMIC_ACQ_REL)){
                /* Remember the destructor.  */
                sys::__internal::__pthread_keys[cnt].destr = destructor;

                /* Return the key to the caller.  */
                *key = cnt;

                /* The call succeeded.  */
                return 0;
            }
        }

        return EAGAIN;
    }

    int pthread_key_delete(pthread_key_t key) {
        int result = EINVAL;

        if (__builtin_expect((key < PTHREAD_KEYS_MAX),1)){
            uintptr_t seq = sys::__internal::__pthread_keys[key].seq;

            if (__builtin_expect(!(((seq) & 1) == 0), 1)
                && !__atomic_compare_exchange(&sys::__internal::__pthread_keys[key].seq,
                    &seq + 1, &seq,false, __ATOMIC_ACQ_REL, __ATOMIC_ACQ_REL))
                /* We deleted a valid key.  */
                result = 0;
        }

        return result;
    }

    void* pthread_getspecific(pthread_key_t key) {
        //struct pthread_key_data* data;

        ///* Special case access to the first 2nd-level block.  This is the
        //   usual case.  */
        //if (__builtin_expect((key < 32),1))
        //    data = &({ class sys::thread* __self; asm("mov %%fs:%c1,%0" : "=r" (__self) : "i" (offsetof(class sys::thread, header.self))); __self; })->specific_1stblock[key];
        //else
        //{
        //    /* Verify the key is sane.  */
        //    if (key >= PTHREAD_KEYS_MAX)
        //        /* Not valid.  */
        //        return NULL;

        //    unsigned int idx1st = key / 32;
        //    unsigned int idx2nd = key % 32;

        //    /* If the sequence number doesn't match or the key cannot be defined
        //   for this thread since the second level array is not allocated
        //   return NULL, too.  */
        //    struct pthread_key_data* level2 = THREAD_GETMEM_NC(THREAD_SELF,
        //        specific, idx1st);
        //    if (level2 == NULL)
        //        /* Not allocated, therefore no data.  */
        //        return NULL;

        //    /* There is data.  */
        //    data = &level2[idx2nd];
        //}

        //void* result = data->data;
        //if (result != NULL)
        //{
        //    uintptr_t seq = data->seq;

        //    if (__builtin_expect((seq != sys::__internal::__pthread_keys[key].seq),0))
        //        result = data->data = NULL;
        //}

        //return result;
    }

    int pthread_getattr_np(pthread_t thread, pthread_attr_t *attr){

    }

    int pthread_setspecific(pthread_key_t key, const void* tls) {

    }

    int pthread_attr_init(pthread_attr_t* args) {
        args = new pthread_attr_t;
    };

    int pthread_attr_destroy(pthread_attr_t* args) {
        delete args;
        return 0;
    }

    int pthread_attr_setstack(pthread_attr_t* attr,
        void* stackaddr, size_t stacksize) {
        if (attr) {
            try {
                ((sys::threadattr*)*attr)->setStackSize(stacksize);
                ((sys::threadattr*)*attr)->setStack(stackaddr);
            }
            catch (...) {
                return -1;
            }
            return 0;
        }
        return -1;
    };

    int pthread_attr_getstack(const __restrict__ pthread_attr_t* attr,
        void** __restrict__ stackaddr,
        size_t* __restrict__ stacksize) {
        if (attr) {
            try {
                ((sys::threadattr*)*attr)->getStack(*stackaddr, *stacksize);
            }
            catch (...) {
                return -1;
            }
            return 0;
        }
        return -1;
    };

    int pthread_attr_setstacksize(pthread_attr_t *attr, size_t stacksize){
            try {
                if(((sys::threadattr*)*attr)){
                    ((sys::threadattr*)*attr)->setStackSize(stacksize);
                    return 0;
                }else{
                    return -1;
                }
            }catch(...){
                return -1;
            }
    }

    int pthread_attr_getstacksize(const pthread_attr_t *__restrict__ attr,
                                     size_t *__restrict__ stacksize){
            try {
                if(((sys::threadattr*)*attr)){
                    *stacksize=((sys::threadattr*)*attr)->getStackSize();
                    return 0;
                }else{
                    return -1;
                }
            }catch(...){
                return -1;
            }
    }

    int pthread_join(pthread_t thread, void** retval) {
        try {
            if (retval)
                *retval = ((sys::posixthread*)thread)->join();
            delete (sys::posixthread*)thread;
            thread = nullptr;
            return 0;
        }
        catch (...) {
            return -1;
        }
    }

    int pthread_create(pthread_t* thread,
        const pthread_attr_t* __restrict__ attr,
        void* (*start_routine)(void*),
        void* __restrict__ arg) {
        try {
            if (attr)
                *thread = new sys::posixthread(*attr, start_routine, arg);
            else
                *thread = new sys::posixthread(nullptr, start_routine, arg);
            return 0;
        }
        catch (...) {
            return -1;
        }
    }

    _Noreturn void pthread_exit(void* retval) {
        _Exit(0);
    };

};


// extern "C" {
//     void *__tls_get_addr(size_t *v){
//         return __threadinternal::getins().getDesc(*v);
//     }
// };
